import numpy as np
import matplotlib as mpl
mpl.use('TkAgg')
from matplotlib import pyplot as plt

from pypulseq.Sequence import block
from pypulseq.Sequence.read_seq import read
from pypulseq.Sequence.write_seq import write
from pypulseq.calc_duration import calc_duration
from pypulseq.event_lib import EventLibrary
from pypulseq.opts import Opts


class Sequence:
    def __init__(self, system=Opts()):
        self.system = system
        # EventLibrary.data is a dict
        self.shape_library = EventLibrary()
        self.rf_library = EventLibrary()
        self.grad_library = EventLibrary()
        self.adc_library = EventLibrary()
        self.delay_library = EventLibrary()
        self.block_events = {}
        self.rf_raster_time = self.system.rf_raster_time
        self.grad_raster_time = self.system.grad_raster_time

    def __str__(self):
        s = "Sequence:"
        s += "\nshape_library: " + str(self.shape_library)
        s += "\nrf_library: " + str(self.rf_library)
        s += "\ngrad_library: " + str(self.grad_library)
        s += "\nadc_library: " + str(self.adc_library)
        s += "\ndelay_library: " + str(self.delay_library)
        s += "\nrf_raster_time: " + str(self.rf_raster_time)
        s += "\ngrad_raster_time: " + str(self.grad_raster_time)
        s += "\nblock_events: " + str(len(self.block_events))
        return s

    def duration(self):
        num_blocks = len(self.block_events)
        event_count = 0
        duration = 0
        for i in range(1, num_blocks + 1):
            block = self.get_block(i)
            duration += calc_duration(block)

        return duration, num_blocks, event_count

    def add_block(self, *args):
        """
        Adds supplied list of Holder objects as a Block.

        Parameters
        ----------
        args : list
            List of Holder objects to be added as a Block.
        """

        block.add_block(self, len(self.block_events) + 1, *args)

    def get_block(self, block_index):
        """
        Returns Block at block_index.

        Parameters
        ----------
        block_index : int
            Index of required block.
        """

        return block.get_block(self, block_index)

    def read(self, file_path):
        read(self, file_path)

    def write(self, name):
        write(self, name)

    def plot(self, time_range=(0, np.inf)):
        """
        Show Matplotlib plot of all Events in the Sequence object.

        Parameters
        ----------
        time_range : List
            Time range (x-axis limits) for plot to be shown. Default is 0 to infinity (entire plot shown).
        """

        fig1, fig2 = plt.figure(1), plt.figure(2)
        f11, f12, f13 = fig1.add_subplot(311), fig1.add_subplot(312), fig1.add_subplot(313)
        f2 = [fig2.add_subplot(311), fig2.add_subplot(312), fig2.add_subplot(313)]
        t0 = 0
        for iB in range(1, len(self.block_events) + 1):
            block = self.get_block(iB)
            is_valid = time_range[0] <= t0 <= time_range[1]
            if is_valid:
                if block is not None:
                    if 'adc' in block:
                        adc = block['adc']
                        t = adc.delay + [(x * adc.dwell) for x in range(0, int(adc.num_samples))]
                        f11.plot((t0 + t), np.zeros(len(t)))
                    if 'rf' in block:
                        rf = block['rf']
                        t = rf.t
                        f12.plot(np.squeeze(t0 + t), abs(rf.signal))
                        f13.plot(np.squeeze(t0 + t), np.angle(rf.signal))
                    grad_channels = ['gx', 'gy', 'gz']
                    for x in range(0, len(grad_channels)):
                        if grad_channels[x] in block:
                            grad = block[grad_channels[x]]
                            if grad.type == 'grad':
                                t = grad.t
                                waveform = 1e-3 * grad.waveform
                            else:
                                t = np.cumsum([0, grad.rise_time, grad.flat_time, grad.fall_time])
                                waveform = [1e-3 * grad.amplitude * x for x in [0, 1, 1, 0]]
                            f2[x].plot(np.squeeze(t0 + t), waveform)
            t0 += calc_duration(block)

        f11.set_ylabel('adc')
        f12.set_ylabel('rf mag hz')
        f13.set_ylabel('rf phase rad')
        [f2[x].set_ylabel(grad_channels[x]) for x in range(3)]
        # Setting display limits
        disp_range = [time_range[0], min(t0, time_range[1])]
        f11.set_xlim(disp_range)
        f12.set_xlim(disp_range)
        f13.set_xlim(disp_range)
        [x.set_xlim(disp_range) for x in f2]

        plt.show()



import numpy as np

from pypulseq.event_lib import EventLibrary


def read(self, path):
    """
    Reads .seq file from path, and constructs a Sequence object from the file.

    Parameters
    ----------
    path : str
        Path of .seq file to be read.
    """

    input_file = open(path, 'r')
    self.shape_library = EventLibrary()
    self.rf_library = EventLibrary()
    self.grad_library = EventLibrary()
    self.adc_library = EventLibrary()
    self.delay_library = EventLibrary()
    self.block_events = {}
    self.rf_raster_time = self.system.rf_raster_time
    self.grad_raster_time = self.system.grad_raster_time

    while True:
        section = skip_comments(input_file)
        if section == -1:
            break
        if section == '[BLOCKS]':
            self.block_events = read_blocks(input_file)
        elif section == '[RF]':
            self.rf_library = read_events(input_file, 1, None, None)
        elif section == '[GRAD]':
            self.grad_library = read_events(input_file, 1, 'g', self.grad_library)
        elif section == '[TRAP]':
            self.grad_library = read_events(input_file, [1, 1e-6, 1e-6, 1e-6], 't', self.grad_library)
        elif section == '[ADC]':
            self.adc_library = read_events(input_file, [1, 1e-9, 1e-6, 1, 1], None, None)
        elif section == '[DELAYS]':
            self.delay_library = read_events(input_file, 1e-6, None, None)
        elif section == '[SHAPES]':
            self.shape_library = read_shapes(input_file)


def read_blocks(input_file):
    """
    Read Blocks from .seq file. Blocks are single lines under the '[BLOCKS]' header in the .seq file.

    Parameters
    ----------
    input_file : file
        .seq file to be read.

    Returns
    -------
    block_events : dict
        Key-value mapping of Block ID and Event ID.
    """

    line = strip_line(input_file)
    for x in range(len(line)):
        line[x] = float(line[x])

    event_table = []
    while not (line == '\n' or line[0] == '#'):
        event_row = []
        for c in line[1:]:
            event_row.append(float(c))
        event_table.append(event_row)

        line = strip_line(input_file)
        # Break here to avoid crash when the while loop condition is evaluated for line != '\n'
        # Crash occurs because spaces have been eliminated
        if len(line) == 0:
            break

    block_events = {}
    for x in range(len(event_table)):
        block_events[x + 1] = np.array(event_table[x])

    return block_events


def read_events(input_file, scale, type, event_lib):
    scale = 1 if scale is None else scale
    event_library = event_lib if event_lib is not None else EventLibrary()

    line = strip_line(input_file)
    for x in range(len(line)):
        line[x] = float(line[x])

    while not (line == '\n' or line[0] == '#'):
        event_id = line[0]
        data = np.multiply(line[1:], scale)
        event_library.insert(event_id, data, type)

        line = strip_line(input_file)
        if not line:
            break

        for x in range(len(line)):
            line[x] = float(line[x])

    return event_library


def read_shapes(input_file):
    shape_library = EventLibrary()

    strip_line(input_file)
    line = strip_line(input_file)

    while not (line == -1 or len(line) == 0 or line[0] != 'shape_id'):
        id = int(line[1])
        line = skip_comments(input_file)
        num_samples = int(line.split(' ')[1])
        data = []
        line = skip_comments(input_file)
        line = line.split(' ')
        while not (len(line) == 0 or line[0] == '#'):
            data.append(float(line[0]))
            line = strip_line(input_file)
        line = skip_comments(input_file)
        # line could be -1 since -1 is EOF marker, returned from skipComments(inputFile)
        line = line.split(' ') if line != -1 else line
        data.insert(0, num_samples)
        data = np.reshape(data, [1, len(data)])
        shape_library.insert(id, data, None)

    return shape_library


def skip_comments(input_file):
    """
    Skip one '#' comment in .seq file.

    Parameters
    ----------
    input_file : file
        .seq file to be read.

    Returns
    -------
    line : str
        First line in input_file after skipping one '#' comment block.
        Note: File pointer is remembered, so successive calls work as expected.
    """

    line = input_file.readline()
    if line == '':
        return -1
    while line == '\n' or line[0] == '#':
        line = input_file.readline()
        if line == '':
            return -1
    line = line.strip()
    return line


def strip_line(input_file):
    """
    Remove spaces, newline whitespace and return line.

    Parameters
    ----------
    input_file : file
        .seq file to be read.

    Returns
    -------
    line : str
        First line in input_file after removing spaces and newline whitespaces.
        Note: File pointer is remembered, so successive calls work as expected.
    """
    line = input_file.readline()
    line = line.strip()
    line = line.split(' ')
    while '' in line:
        line.remove('')
    return line



from decimal import *

from pypulseq.holder import Holder
from pypulseq.opts import Opts


def makeadc(kwargs):
    """
    Makes a Holder object for an ADC Event.

    Parameters
    ----------
    kwargs : dict
        Key value mappings of ADC Event parameters_params and values.

    Returns
    -------
    adc : Holder
        ADC Event.
    """

    num_samples = kwargs.get("num_samples", 0)
    system = kwargs.get("system", Opts())
    dwell = kwargs.get("dwell", 0)
    duration = kwargs.get("duration", 0)
    delay = kwargs.get("delay", 0)
    freq_offset = kwargs.get("freq_offset", 0)
    phase_offset = kwargs.get("phase_offset", 0)

    adc = Holder()
    adc.type = 'adc'
    adc.num_samples = num_samples
    adc.dwell = dwell
    adc.delay = delay
    adc.freq_offset = freq_offset
    adc.phase_offset = phase_offset
    adc.dead_time = system.adc_dead_time

    if (dwell == 0 and duration == 0) or (dwell > 0 and duration > 0):
        raise ValueError("Either dwell or duration must be defined, not both")

    if duration > 0:
        # getcontext().prec = 4
        adc.dwell = float(Decimal(duration) / Decimal(num_samples))
    adc.duration = dwell * num_samples if dwell > 0 else 0
    return adc



import numpy as np

from pypulseq.compress_shape import compress_shape
from pypulseq.decompress_shape import decompress_shape
from pypulseq.holder import Holder


def add_block(self, block_index, *args):
    """
    Adds supplied list of Holder objects at position specified by block_index.

    Parameters
    ----------
    block_index : int
        Index at which Block has to be inserted.
    args : list
        List of Holder objects to be added as a Block.
    """

    self.block_events[block_index] = np.zeros(6)
    duration = 0
    for event in args:
        if event.type == 'rf':
            mag = np.abs(event.signal)
            amplitude = np.max(mag)
            mag = np.divide(mag, amplitude)
            # Following two lines of code are workarounds for numpy's divide functions returning NaN when mathematical
            # edge cases are encountered (eg. divide by 0)
            mag[np.isnan(mag)] = 0
            phase = np.angle(event.signal)
            phase[np.where(phase < 0)] += 2 * np.pi
            phase /= 2 * np.pi

            mag_shape = compress_shape(mag)
            # data = np.array([[mag_shape.num_samples]])
            # data = np.append(data, mag_shape.data, axis=1)
            data = np.hstack((mag_shape.num_samples, np.ravel(mag_shape.data))).reshape((1, -1))
            mag_id, found = self.shape_library.find(data)
            if not found:
                self.shape_library.insert(mag_id, data, None)

            phase_shape = compress_shape(phase)
            data = np.hstack((phase_shape.num_samples, np.ravel(phase_shape.data))).reshape((1, -1))
            phase_id, found = self.shape_library.find(data)
            if not found:
                self.shape_library.insert(phase_id, data, None)

            data = np.array([amplitude, mag_id, phase_id, event.freq_offset, event.phase_offset, event.dead_time,
                             event.ring_down_time])
            data_id, found = self.rf_library.find(data)
            if not found:
                self.rf_library.insert(data_id, data, None)

            self.block_events[block_index][1] = data_id
            duration = max(duration, max(mag.shape) * self.rf_raster_time + event.dead_time + event.ring_down_time)
        elif event.type == 'grad':
            channel_num = ['x', 'y', 'z'].index(event.channel)
            amplitude = max(abs(event.waveform[0]))
            g = event.waveform / amplitude
            shape = compress_shape(g)
            data = np.array([[shape.num_samples]])
            data = np.append(data, shape.data, axis=1)
            shape_id, found = self.shape_library.find(data)
            if not found:
                self.shape_library.insert(shape_id, data, None)
            data = np.array([amplitude, shape_id])
            index, found = self.grad_library.find(data)
            if not found:
                self.grad_library.insert(index, data, 'g')
            idx = 2 + channel_num
            self.block_events[block_index][idx] = index
            duration = max(duration, len(g[0]) * self.grad_raster_time)
        elif event.type == 'trap':
            channel_num = ['x', 'y', 'z'].index(event.channel)
            data = np.array([event.amplitude, event.rise_time, event.flat_time, event.fall_time])
            index, found = self.grad_library.find(data)
            if not found:
                self.grad_library.insert(index, data, 't')
            idx = 2 + channel_num
            self.block_events[block_index][idx] = index
            duration = max(duration, event.rise_time + event.flat_time + event.fall_time)
        elif event.type == 'adc':
            data = np.array(
                [event.num_samples, event.dwell, event.delay, event.freq_offset, event.phase_offset, event.dead_time])
            index, found = self.adc_library.find(data)
            if not found:
                self.adc_library.insert(index, data, None)
            self.block_events[block_index][5] = index
            duration = max(duration, event.delay + event.num_samples * event.dwell + event.dead_time)
        elif event.type == 'delay':
            data = np.array([event.delay])
            index, found = self.delay_library.find(data)
            if not found:
                self.delay_library.insert(index, data, None)
            self.block_events[block_index][0] = index
            duration = max(duration, event.delay)


def get_block(self, block_index):
    """
    Returns Block at position specified by block_index.

    Parameters
    ----------
    block_index : int
        Index of Block to be retrieved.

    Returns
    -------
    block : dict
        Block at position specified by block_index.
    """

    block = {}
    event_ind = self.block_events[block_index]
    if event_ind[0] > 0:
        delay = Holder()
        delay.type = 'delay'
        delay.delay = self.delay_library.data[event_ind[0]]
        block['delay'] = delay
    elif event_ind[1] > 0:
        rf = Holder()
        rf.type = 'rf'
        lib_data = self.rf_library.data[event_ind[1]]

        amplitude, mag_shape, phase_shape = lib_data[0], lib_data[1], lib_data[2]
        shape_data = self.shape_library.data[mag_shape]
        compressed = Holder()
        compressed.num_samples = shape_data[0][0]
        compressed.data = shape_data[0][1:]
        compressed.data = compressed.data.reshape((1, compressed.data.shape[0]))
        mag = decompress_shape(compressed)
        shape_data = self.shape_library.data[phase_shape]
        compressed.num_samples = shape_data[0][0]
        compressed.data = shape_data[0][1:]
        compressed.data = compressed.data.reshape((1, compressed.data.shape[0]))
        phase = decompress_shape(compressed)
        rf.signal = 1j * 2 * np.pi * phase
        rf.signal = amplitude * mag * np.exp(rf.signal)
        rf.t = [(x * self.rf_raster_time) for x in range(1, max(mag.shape) + 1)]
        rf.t = np.reshape(rf.t, (1, len(rf.t)))
        rf.freq_offset = lib_data[3]
        rf.phase_offset = lib_data[4]
        if max(lib_data.shape) < 6:
            lib_data = np.append(lib_data, 0)
        rf.dead_time = lib_data[5]

        if max(lib_data.shape) < 7:
            lib_data = np.append(lib_data, 0)
        rf.ring_down_time = lib_data[6]

        block['rf'] = rf
    grad_channels = ['gx', 'gy', 'gz']
    for i in range(1, len(grad_channels) + 1):
        if event_ind[2 + (i - 1)] > 0:
            grad, compressed = Holder(), Holder()
            type = self.grad_library.type[event_ind[2 + (i - 1)]]
            lib_data = self.grad_library.data[event_ind[2 + (i - 1)]]
            grad.type = 'trap' if type == 't' else 'grad'
            grad.channel = grad_channels[i - 1][1]
            if grad.type == 'grad':
                amplitude = lib_data[0]
                shape_id = lib_data[1]
                shape_data = self.shape_library.data[shape_id]
                compressed.num_samples = shape_data[0][0]
                compressed.data = np.array([shape_data[0][1:]])
                g = decompress_shape(compressed)
                grad.waveform = amplitude * g
                grad.t = np.array([[x * self.grad_raster_time for x in range(1, g.size + 1)]])
            else:
                grad.amplitude, grad.rise_time, grad.flat_time, grad.fall_time = [lib_data[x] for x in range(4)]
                grad.area = grad.amplitude * (grad.flat_time + grad.rise_time / 2 + grad.fall_time / 2)
                grad.flat_area = grad.amplitude * grad.flat_time
            block[grad_channels[i - 1]] = grad

    if event_ind[5] > 0:
        lib_data = self.adc_library.data[event_ind[5]]
        if max(lib_data.shape) < 6:
            lib_data = np.append(lib_data, 0)
        adc = Holder()
        adc.num_samples, adc.dwell, adc.delay, adc.freq_offset, adc.phase_offset, adc.dead_time = [lib_data[x] for x in
                                                                                                   range(6)]
        adc.type = 'adc'
        block['adc'] = adc
    return block



from math import pi


def convert_from_to(from_value, from_unit):
    if isinstance(from_value, float) and isinstance(from_unit, str):
        gamma, standard = 42.576e6, 0
        # Converting gradient
        if from_unit == 'Hz/m':
            standard = from_value
        elif from_unit == 'mT/m':
            standard = from_value * 1e-3 * gamma
        elif from_unit == 'rad/ms/mm':
            standard = from_value * 1e6 / (2 * pi)
        # Converting slew rate
        elif from_unit == 'Hz/m/s':
            standard = from_value
        elif from_unit == 'mT/m/ms' or 'T/m/s':
            standard = from_value * gamma
        elif from_unit == 'rad/ms/mm/ms':
            standard = from_value * 1e9 / (2 * pi)
        return standard
    else:
        raise TypeError("input parameters_params should be: from_value, from_unit")



import numpy as np


def write(self, file_name):
    """
    Writes a .seq file from the Sequence object calling the method.
    >.0f is used when only decimals have to be displayed.
    >g is used when insignificant zeros have to be truncated.

    Parameters
    ----------
    file_name : str
        File name of .seq file to be written.
    """
    file_name += '.seq' if file_name[-4:] != '.seq' not in file_name else ''
    output_file = open(file_name, 'w')
    output_file.write("# Pulseq sequence file\n")
    output_file.write("# Created by Python/GPI Lab\n\n")
    output_file.write("# Format of blocks:\n")
    output_file.write("#  #  D RF  GX  GY  GZ ADC\n")
    output_file.write("[BLOCKS]\n")
    id_format_width = len(str(len(self.block_events)))
    id_format_str = '{:>' + str(id_format_width) + '}'
    id_format_str += ' {:>2.0f} {:>2.0f} {:>3.0f} {:>3.0f} {:>3.0f} {:>2.0f}\n'
    for i in range(0, len(self.block_events)):
        s = id_format_str.format(*np.insert(self.block_events[i + 1].astype(int), 0, (i + 1)))
        output_file.write(s)
    output_file.write('\n')

    if len(self.rf_library.keys) != 0:
        output_file.write('# Format of RF events:\n')
        output_file.write('# id amplitude mag_id phase_id freq phase\n')
        output_file.write('# ..        Hz   ....     ....   Hz   rad\n')
        output_file.write('[RF]\n')
        rf_lib_keys = self.rf_library.keys
        id_format_str = '{:>1.0f} {:>12g} {:>1.0f} {:>1.0f} {:>g} {:>g}\n'
        for k in rf_lib_keys.keys():
            lib_data = self.rf_library.data[k][0:5]
            s = id_format_str.format(*np.insert(lib_data, 0, k))
            output_file.write(s)
        output_file.write('\n')

    grad_lib_values = np.array(list(self.grad_library.type.values()))
    arb_grad_mask = np.where(grad_lib_values == 'g')[0]
    trap_grad_mask = np.where(grad_lib_values == 't')[0]

    if any(arb_grad_mask):
        output_file.write('# Format of arbitrary gradients:\n')
        output_file.write('# id amplitude shape_id\n')
        output_file.write('# ..      Hz/m     ....\n')
        output_file.write('[GRADIENTS]\n')
        keys = np.array(list(self.grad_library.keys.keys()))
        id_format_str = '{:>1.0f} {:>12g} {:>1.0f} \n'
        for k in keys[arb_grad_mask]:
            s = id_format_str.format(*np.insert(self.grad_library.data[k], 0, k))
            output_file.write(s)
        output_file.write('\n')

    if any(trap_grad_mask):
        output_file.write('# Format of trapezoid gradients:\n')
        output_file.write('# id amplitude rise flat fall\n')
        output_file.write('# ..      Hz/m   us   us   us\n')
        output_file.write('[TRAP]\n')
        keys = np.array(list(self.grad_library.keys.keys()))
        for k in keys[trap_grad_mask]:
            id_format_str = '{:>2.0f} {:>12g} {:>3.0f} {:>4.0f} {:>3.0f}\n'
            data = self.grad_library.data[k]
            data = np.reshape(data, (1, data.shape[0]))
            data[0][1:] = np.round(1e6 * data[0][1:], decimals=2)
            s = id_format_str.format(*np.insert(data, 0, k))
            output_file.write(s)
        output_file.write('\n')

    if len(self.adc_library.keys) != 0:
        output_file.write('# Format of ADC events:\n')
        output_file.write('# id num dwell delay freq phase\n')
        output_file.write('# ..  ..    ns    us   Hz   rad\n')
        output_file.write('[ADC]\n')
        keys = self.adc_library.keys
        id_format_str = '{:>2.0f} {:>3.0f} {:>6.0f} {:>3.0f} {:>g} {:>g}\n'
        for k in keys.values():
            data = np.multiply(self.adc_library.data[k][0:5], [1, 1e9, 1e6, 1, 1])
            s = id_format_str.format(*np.insert(data, 0, k))
            output_file.write(s)
    output_file.write('\n')

    if len(self.delay_library.keys) != 0:
        output_file.write('# Format of delays:\n')
        output_file.write('# id delay (us)\n')
        output_file.write('[DELAYS]\n')
        keys = self.delay_library.keys
        id_format_str = '{:>.0f} {:>.0f}\n'
        for k in keys.values():
            data = np.round(1e6 * self.delay_library.data[k])
            s = id_format_str.format(*np.insert(data, 0, k))
            output_file.write(s)
        output_file.write('\n')

    if len(self.shape_library.keys) != 0:
        output_file.write('# Sequence Shapes\n')
        output_file.write('[SHAPES]\n\n')
        keys = self.shape_library.keys
        for k in keys.values():
            shape_data = self.shape_library.data[k]
            s = 'shape_id {:>.0f}\n'
            s = s.format(k)
            output_file.write(s)
            s = 'num_samples {:>g}\n'
            s = s.format(shape_data[0][0])
            output_file.write(s)
            s = '{:g}\n'
            for x in shape_data[0][1:]:
                s1 = s.format(x)
                output_file.write(s1)
            output_file.write('\n')



from pypulseq.holder import Holder


def calc_duration(*events):
    """
    Calculate the cumulative duration of Events.

    Parameters
    ----------
    events : list
        List of Holder objects. Can also be a list containing a single block (see plot() in Sequence).

    Returns
    -------
    duration : float
        The cumulative duration of the Events passed.
    """
    if isinstance(events[0], dict):
        events = events[0].values()

    duration = 0
    for event in events:
        if not isinstance(event, Holder):
            raise TypeError("input(s) should be of type core.events.Holder")

        if event.type == 'delay':
            duration = max(duration, event.delay)
        elif event.type == 'rf':
            duration = max(duration, event.t[0][-1] + event.dead_time + event.ring_down_time)
        elif event.type == 'grad':
            duration = max(duration, event.t[0][-1])
        elif event.type == 'adc':
            adc_time = event.delay + event.num_samples * event.dwell + event.dead_time
            duration = max(duration, adc_time)
        elif event.type == 'trap':
            duration = max(duration, event.rise_time + event.flat_time + event.fall_time)

    return duration



from math import pi

import numpy as np

from pypulseq.holder import Holder
from pypulseq.make_trap import make_trapezoid
from pypulseq.opts import Opts


def make_arbitrary_rf(kwargs, nargout=1):
    """
    Makes a Holder object for an arbitrary RF pulse Event.

    Parameters
    ----------
    kwargs : dict
        Key value mappings of RF Event parameters_params and values.
    nargout: int
        Number of output arguments to be returned. Default is 1, only RF Event is returned. Passing any number greater
        than 1 will return the Gz Event along with the RF Event.

    Returns
    -------
    rf : Holder
        RF Event configured based on supplied kwargs.
    gz : Holder
        Slice select trapezoidal gradient Event.
    """

    flip_angle = kwargs.get("flip_angle")
    system = kwargs.get("system", Opts())
    signal = kwargs.get("signal", 0)
    freq_offset = kwargs.get("freq_offset", 0)
    phase_offset = kwargs.get("phase_offset", 0)
    time_bw_product = kwargs.get("time_bw_product", 0)
    bandwidth = kwargs.get("bandwidth", 0)
    max_grad = kwargs.get("max_grad", 0)
    max_slew = kwargs.get("max_slew", 0)
    slice_thickness = kwargs.get("slice_thickness", 0)

    signal = signal / sum(signal * system.rf_raster_time) * flip_angle / (2 * pi)
    N = len(signal)
    duration = N * system.rf_raster_time
    t = [x * system.rfRasterTime for x in range(0, N)]

    rf = Holder()
    rf.type = 'rf'
    rf.signal = signal
    rf.t = t
    rf.freq_offset = freq_offset
    rf.phase_offset = phase_offset
    rf.rf_dead_time = system.rf_dead_time
    rf.ring_down_time = system.rf_ring_down_time

    if nargout > 1:
        if slice_thickness <= 0:
            raise ValueError('Slice thickness must be provided')
        if bandwidth <= 0:
            raise ValueError('Bandwidth of pulse must be provided')

        system.maxgrad = max_grad if max_grad > 0 else system.maxgrad
        system.max_slew = max_slew if max_slew > 0 else system.max_slew

        BW = bandwidth
        if time_bw_product > 0:
            BW = time_bw_product / duration

        amplitude = BW / slice_thickness
        area = amplitude * duration
        kwargs_for_trap = {"channel": 'z', "system": system, "flat_time": duration, "flat_area": area}
        gz = make_trapezoid(**kwargs_for_trap)

        t_fill = np.arange(1, round(gz.rise_time / 1e-6) + 1) * 1e-6
        rf.t = [t_fill, rf.t + t_fill[-1], t_fill + rf.t[-1] + t_fill[-1]]
        rf.signal = [np.zeros(len(t_fill)), rf.signal, np.zeros(len(t_fill))]

    if rf.ring_down_time > 0:
        t_fill = np.arange(1, round(gz.rise_time / 1e-6) + 1) * 1e-6
        rf.t = [rf.t, rf.t[-1] + t_fill]
        rf.signal = [rf.signal, np.zeros(len(t_fill))]

    return rf, gz



import pypulseq.convert as convert


class Opts():
    """This class contains the gradient limits of the MR system."""

    def __init__(self, kwargs=dict()):
        valid_grad_units = ['Hz/m', 'mT/m', 'rad/ms/mm']
        valid_slew_units = ['Hz/m/s', 'mT/m/ms', 'T/m/s', 'rad/ms/mm/ms']
        self.max_grad = kwargs.get("max_grad", 40)
        self.max_slew = kwargs.get("max_slew", 170)
        self.grad_unit = kwargs.get("grad_unit", valid_grad_units[1])
        self.slew_unit = kwargs.get("slew_unit", valid_slew_units[1])

        # Convert input values if not provided in standard units
        self.max_grad = convert.convert_from_to(float(self.max_grad), self.grad_unit)
        self.max_slew = convert.convert_from_to(float(self.max_slew), self.slew_unit)

        self.te = kwargs.get("te", 0)
        self.tr = kwargs.get("tr", 0)
        self.flip = kwargs.get("flip", 0)
        self.fov = kwargs.get("fov", 0)
        self.Nx = kwargs.get("Nx", 0)
        self.Ny = kwargs.get("Ny", 0)
        self.rise_time = kwargs.get("rise_time", 0)
        self.rf_dead_time = kwargs.get("rf_dead_time", 0)
        self.rf_raster_time = kwargs.get("rf_raster_time", 1e-6)
        self.rf_ring_down_time = kwargs.get("rf_ring_down_time", 0)
        self.adc_dead_time = kwargs.get("adc_dead_time", 0)
        self.grad_raster_time = kwargs.get("grad_raster_time", 10e-6)

    def __str__(self):
        s = "System limits:"
        s += "\nmax_grad: " + str(self.max_grad) + str(self.grad_unit)
        s += "\nmax_slew: " + str(self.max_slew) + str(self.slew_unit)
        s += "\nte: " + str(self.te)
        s += "\ntr: " + str(self.tr)
        s += "\nflip: " + str(self.flip)
        s += "\nfov: " + str(self.fov)
        s += "\nNx: " + str(self.Nx)
        s += "\nNy: " + str(self.Ny)
        s += "\nrise_time: " + str(self.rise_time)
        s += "\nrf_dead_time: " + str(self.rf_dead_time)
        s += "\nadc_dead_time: " + str(self.adc_dead_time)
        s += "\nrf_raster_time: " + str(self.rf_raster_time)
        s += "\ngrad_raster_time: " + str(self.grad_raster_time)
        return s



from pypulseq.holder import Holder


def make_delay(d):
    """
    Makes a Holder object for an delay Event.

    Parameters
    ----------
    d : float
        Delay time, in seconds.

    Returns
    -------
    delay : Holder
        Delay Event.
    """

    delay = Holder()
    if d < 0:
        raise ValueError('Delay {:.2f} ms is invalid'.format(d * 1e3))
    delay.type = 'delay'
    delay.delay = d
    return delay



import numpy as np

from pypulseq.holder import Holder
from pypulseq.make_trap import make_trapezoid
from pypulseq.opts import Opts


def make_sinc_pulse(kwargs, nargout=1):
    """
    Makes a Holder object for an RF pulse Event.

    Parameters
    ----------
    kwargs : dict
        Key value mappings of RF Event parameters_params and values.
    nargout: int
        Number of output arguments to be returned. Default is 1, only RF Event is returned. Passing any number greater
        than 1 will return the Gz Event along with the RF Event.

    Returns
    -------
    rf : Holder
        RF Event configured based on supplied kwargs.
    gz : Holder
        Slice select trapezoidal gradient Event.
    """

    flip_angle = kwargs.get("flip_angle")
    system = kwargs.get("system", Opts())
    duration = kwargs.get("duration", 0)
    freq_offset = kwargs.get("freq_offset", 0)
    phase_offset = kwargs.get("phase_offset", 0)
    time_bw_product = kwargs.get("time_bw_product", 4)
    apodization = kwargs.get("apodization", 0)
    max_grad = kwargs.get("max_grad", 0)
    max_slew = kwargs.get("max_slew", 0)
    slice_thickness = kwargs.get("slice_thickness", 0)

    BW = time_bw_product / duration
    alpha = apodization
    N = int(round(duration / 1e-6))
    t = np.zeros((1, N))
    for x in range(1, N + 1):
        t[0][x - 1] = x * system.rf_raster_time
    tt = t - (duration / 2)
    window = np.zeros((1, tt.shape[1]))
    for x in range(0, tt.shape[1]):
        window[0][x] = 1.0 - alpha + alpha * np.cos(2 * np.pi * tt[0][x] / duration)
    signal = np.multiply(window, np.sinc(BW * tt))
    flip = np.sum(signal) * system.rf_raster_time * 2 * np.pi
    signal = signal * flip_angle / flip

    rf = Holder()
    rf.type = 'rf'
    rf.signal = signal
    rf.t = t
    rf.freq_offset = freq_offset
    rf.phase_offset = phase_offset
    rf.dead_time = system.rf_dead_time
    rf.ring_down_time = system.rf_ring_down_time

    fill_time = 0
    if nargout > 1:
        if slice_thickness == 0:
            raise ValueError('Slice thickness must be provided')

        system.max_grad = max_grad if max_grad > 0 else system.max_grad
        system.max_slew = max_slew if max_slew > 0 else system.max_slew

        amplitude = BW / slice_thickness
        area = amplitude * duration
        kwargs_for_trap = {"channel": 'z', "system": system, "flat_time": duration, "flat_area": area}
        gz = make_trapezoid(kwargs_for_trap)

        fill_time = gz.rise_time
        nfill_time = int(round(fill_time / 1e-6))
        t_fill = np.zeros((1, nfill_time))
        for x in range(1, nfill_time + 1):
            t_fill[0][x - 1] = x * 1e-6
        temp = np.concatenate((t_fill[0], rf.t[0] + t_fill[0][-1]))
        temp = temp.reshape((1, len(temp)))
        rf.t = np.resize(rf.t, temp.shape)
        rf.t[0] = temp
        z = np.zeros((1, t_fill.shape[1]))
        temp2 = np.concatenate((z[0], rf.signal[0]))
        temp2 = temp2.reshape((1, len(temp2)))
        rf.signal = np.resize(rf.signal, temp2.shape)
        rf.signal[0] = temp2

    # Add dead time to start of pulse, if required
    if fill_time < rf.dead_time:
        fill_time = rf.dead_time - fill_time
        t_fill = (np.arange(int(round(fill_time / 1e-6))) * 1e-6)[np.newaxis, :]
        rf.t = np.concatenate((t_fill, (rf.t + t_fill[0, -1])), axis=1)
        rf.signal = np.concatenate((np.zeros(t_fill.shape), rf.signal), axis=1)

    if rf.ring_down_time > 0:
        t_fill = (np.arange(1, round(rf.ring_down_time / 1e-6) + 1) * 1e-6)[np.newaxis, :]
        rf.t = np.concatenate((rf.t, rf.t[0, -1] + t_fill), axis=1)
        rf.signal = np.concatenate((rf.signal, np.zeros(t_fill.shape)), axis=1)

    # Following 2 lines of code are workarounds for numpy returning 3.14... for np.angle(-0.00...)
    negative_zero_indices = np.where(rf.signal == -0.0)
    rf.signal[negative_zero_indices] = 0

    if nargout > 1:
        return rf, gz
    else:
        return rf



import numpy as np


class EventLibrary:
    def __init__(self):
        # size of data is 0x0 because of range_len in find()
        self.keys, self.data, self.lengths, self.type, self.keymap = {}, {}, {}, {}, {}

    def __str__(self):
        s = "EventLibrary:"
        s += "\nkeys: " + str(len(self.keys))
        s += "\ndata: " + str(len(self.data))
        s += "\nlengths: " + str(len(self.lengths))
        s += "\ntype: " + str(len(self.type))
        return s

    def find(self, new_data):
        data_string = np.array2string(new_data, formatter={'float_kind': lambda x: "%.6g" % x})
        data_string = data_string.replace('[', '')
        data_string = data_string.replace(']', '')
        try:
            key_id = self.keymap[data_string]
            found = True
        except:
            key_id = 1 if len(self.keys) == 0 else max(self.keys) + 1
            found = False

        return key_id, found

    def insert(self, key_id, new_data, data_type):
        if not isinstance(new_data, np.ndarray):
            new_data = np.array(new_data)
        self.keys[key_id] = key_id
        self.data[key_id] = new_data
        self.lengths[key_id] = max(self.data[key_id].shape)
        data_string = np.array2string(new_data, formatter={'float_kind': lambda x: "%.6g" % x})
        data_string = data_string.replace('[', '')
        data_string = data_string.replace(']', '')
        self.keymap[data_string] = key_id
        self.type[key_id] = data_type



import numpy as np

from pypulseq.holder import Holder
from pypulseq.make_trap import make_trapezoid
from pypulseq.opts import Opts


def make_block_pulse(kwargs, nargout=1):
    """
    Makes a Holder object for an RF pulse Event.

    Parameters
    ----------
    kwargs : dict
        Key value mappings of RF Event parameters_params and values.
    nargout: int
        Number of output arguments to be returned. Default is 1, only RF Event is returned. Passing any number greater
        than 1 will return the Gz Event along with the RF Event.

    Returns
    -------
    Tuple consisting of:
    rf : Holder
        RF Event configured based on supplied kwargs.
    gz : Holder
        Slice select trapezoidal gradient Event.
    """

    flip_angle = kwargs.get("flip_angle")
    system = kwargs.get("system", Opts())
    duration = kwargs.get("duration", 0)
    freq_offset = kwargs.get("freq_offset", 0)
    phase_offset = kwargs.get("phase_offset", 0)
    time_bw_product = kwargs.get("time_bw_product", 4)
    bandwidth = kwargs.get("bandwidth", 0)
    max_grad = kwargs.get("max_grad", 0)
    max_slew = kwargs.get("max_slew", 0)
    slice_thickness = kwargs.get("slice_thickness", 0)

    if duration == 0:
        if time_bw_product > 0:
            duration = time_bw_product / bandwidth
        elif bandwidth > 0:
            duration = 1 / (4 * bandwidth)
        else:
            raise ValueError('Either bandwidth or duration must be defined')

    BW = 1 / (4 * duration)
    N = round(duration / 1e-6)
    t = [x * system.rf_raster_time for x in range(N)]
    signal = flip_angle / (2 * np.pi) / duration * np.ones(len(t))

    rf = Holder()
    rf.type = 'rf'
    rf.signal = signal
    rf.t = t
    rf.freq_offset = freq_offset
    rf.phase_offset = phase_offset
    rf.dead_time = system.rf_dead_time
    rf.ring_down_time = system.rf_ring_down_time

    fill_time = 0
    if nargout > 1:
        if slice_thickness < 0:
            raise ValueError('Slice thickness must be provided')

        if max_grad > 0:
            system.max_grad = max_grad
        if max_slew > 0:
            system.max_slew = max_slew

        amplitude = BW / slice_thickness
        area = amplitude * duration
        kwargs_for_trap = {'channel': 'z', 'system': system, 'flat_time': duration, 'flat_area': area}
        gz = make_trapezoid(kwargs_for_trap)

        fill_time = gz.rise_time
        t_fill = np.array([x * 1e-6 for x in range(int(round(fill_time / 1e-6)))])
        rf.t = np.array([t_fill, rf.t + t_fill[-1], t_fill + rf.t[-1] + t_fill[-1]])
        rf.signal = np.array([np.zeros(t_fill.size), rf.signal, np.zeros(t_fill.size)])

    if fill_time < rf.dead_time:
        fill_time = rf.dead_time - fill_time
        t_fill = np.array([x * 1e-6 for x in range(int(round(fill_time / 1e-6)))])
        rf.t = np.insert(rf.t, 0, t_fill) + t_fill[-1]
        rf.t = np.reshape(rf.t, (1, len(rf.t)))
        rf.signal = np.insert(rf.signal, 0, np.zeros(t_fill.size))
        rf.signal = np.reshape(rf.signal, (1, len(rf.signal)))

    if rf.ring_down_time > 0:
        t_fill = np.arange(1, round(rf.ring_down_time / 1e-6) + 1) * 1e-6
        rf.t = [rf.t, rf.t[-1] + t_fill]
        rf.signal = [rf.signal, np.zeros(len(t_fill))]

    if nargout > 1:
        return rf, gz
    else:
        return rf



import numpy as np

from pypulseq.holder import Holder


def compress_shape(decompressed_shape):
    """
    Returns a run-length encoded compressed shape.

    Parameters
    ----------
    decompressed_shape : ndarray
        Decompressed shape.

    Returns
    -------
    compressed_shape : Holder
        A Holder object containing the shape of the compressed shape ndarray and the compressed shape ndarray itself.
    """

    if decompressed_shape.shape[0] != 1:
        raise ValueError("input should be of shape (1,x)")
    if not isinstance(decompressed_shape, np.ndarray):
        raise TypeError("input should be of type numpy.ndarray")

    # data = np.array([decompressed_shape[0][0]])
    # data = np.concatenate((data, np.diff(decompressed_shape[0])))
    data = np.hstack((decompressed_shape[0][0], np.diff(decompressed_shape[0])))

    # TODO
    mask_changes = np.hstack((1, np.abs(np.diff(data)) > 1e-8))
    vals = data[np.nonzero(mask_changes)].astype(float)
    k = np.array(np.nonzero(np.append(mask_changes, 1)))
    k = k.reshape((1, k.shape[1]))
    n = np.diff(k)[0]

    n_extra = (n - 2).astype(float)
    vals2 = np.copy(vals)
    vals2[np.where(n_extra < 0)] = np.NAN
    n_extra[np.where(n_extra < 0)] = np.NAN
    v = np.array([vals, vals2, n_extra])
    v = np.concatenate(np.hsplit(v, v.shape[1]))
    finite_vals = np.isfinite(v)
    v = v[finite_vals]
    v_abs = abs(v)
    smallest_indices = np.where(v_abs < 1e-10)
    v[smallest_indices] = 0

    compressed_shape = Holder()
    compressed_shape.num_samples = decompressed_shape.shape[1]
    compressed_shape.data = v.reshape((1, -1))
    return compressed_shape



from math import ceil, sqrt, pow

from pypulseq.holder import Holder
from pypulseq.opts import Opts


def make_trapezoid(kwargs):
    """
    Makes a Holder object for an trapezoidal gradient Event.

    Parameters
    ----------
    kwargs : dict
        Key value mappings of trapezoidal gradient Event parameters_params and values.

    Returns
    -------
    grad : Holder
        Trapezoidal gradient Event configured based on supplied kwargs.
    """

    channel = kwargs.get("channel", "z")
    system = kwargs.get("system", Opts())
    duration = kwargs.get("duration", 0)
    area_result = kwargs.get("area", -1)
    flat_time_result = kwargs.get("flat_time", 0)
    flat_area_result = kwargs.get("flat_area", -1)
    amplitude_result = kwargs.get("amplitude", -1)
    max_grad = kwargs.get("max_grad", 0)
    max_slew = kwargs.get("max_slew", 0)
    rise_time = kwargs.get("rise_time", 0)

    max_grad = max_grad if max_grad > 0 else system.max_grad
    max_slew = max_slew if max_slew > 0 else system.max_slew
    rise_time = rise_time if rise_time > 0 else system.rise_time

    if area_result == -1 and flat_area_result == -1 and amplitude_result == -1:
        raise ValueError('Must supply either ''area'', ''flat_area'' or ''amplitude''')

    if flat_time_result > 0:
        amplitude = amplitude_result if (amplitude_result != -1) else (flat_area_result / flat_time_result)
        if rise_time == 0:
            rise_time = abs(amplitude) / max_slew
            rise_time = ceil(rise_time / system.grad_raster_time) * system.grad_raster_time
        fall_time, flat_time = rise_time, flat_time_result
    elif duration > 0:
        if amplitude_result != -1:
            amplitude = amplitude_result
        else:
            if rise_time == 0:
                dC = 1 / abs(2 * max_slew) + 1 / abs(2 * max_slew)
                amplitude = (duration - sqrt(pow(duration, 2) - 4 * abs(area_result) * dC)) / (2 * dC)
            else:
                amplitude = area_result / (duration - rise_time)

        if rise_time == 0:
            rise_time = ceil(
                amplitude / max_slew / system.grad_raster_time) * system.grad_raster_time

        fall_time = rise_time
        flat_time = (duration - rise_time - fall_time)

        amplitude = area_result / (rise_time / 2 + fall_time / 2 + flat_time) if amplitude_result == -1 else amplitude
    else:
        raise ValueError('Must supply a duration')

    if abs(amplitude) > max_grad:
        raise ValueError("Amplitude violation")

    grad = Holder()
    grad.type = "trap"
    grad.channel = channel
    grad.amplitude = amplitude
    grad.rise_time = rise_time
    grad.flat_time = flat_time
    grad.fall_time = fall_time
    grad.area = amplitude * (flat_time + rise_time / 2 + fall_time / 2)
    grad.flat_area = amplitude * flat_time

    return grad



import numpy as np

from pypulseq.holder import Holder
from pypulseq.opts import Opts


def makearbitrary_grad(kwargs):
    """
    Makes a Holder object for an arbitrary gradient Event.

    Parameters
    ----------
    kwargs : dict
        Key value mappings of RF Event parameters_params and values.

    Returns
    -------
    grad : Holder
        Trapezoidal gradient Event configured based on supplied kwargs.
    """

    channel = kwargs.get("channel", "z")
    system = kwargs.get("system", Opts())
    waveform = kwargs.get("waveform")
    max_grad = kwargs.get("max_grad") if kwargs.get("max_grad", 0) > 0 else system.max_grad
    max_slew = kwargs.get("max_slew ") if kwargs.get("max_slew ", 0) > 0 else system.max_slew

    g = np.reshape(waveform, (1, -1))
    slew = (g[0][1:] - g[0][0:-1]) / system.grad_raster_time
    if max(abs(slew)) > max_slew:
        raise ValueError('Slew rate violation {:f}'.format(max(abs(slew)) / max_slew * 100))
    if max(abs(g[0])) > max_grad:
        raise ValueError('Gradient amplitude violation {:f}'.format(max(abs(g)) / max_grad * 100))
    grad = Holder()
    grad.type = 'grad'
    grad.channel = channel
    grad.waveform = g
    grad.t = np.arange(len(g[0])) * system.grad_raster_time
    grad.t = np.reshape(grad.t, (1, -1))
    return grad



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pypulseq",
    version="1.0.0r1",
    author="Keerthi Sravan Ravi",
    author_email="ks3621@columbia.edu",
    description="Pulseq in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/imr-framework/pypulseq",
    packages=setuptools.find_packages(exclude=['pypulseq.utils', 'pypulseq.examples']),
    install_requires=['numpy', 'matplotlib'],
    license='License :: OSI Approved :: GNU Affero General Public License v3',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ],
)
